package Klava ;

import java.util.Vector ;

public class TupleSpace implements TupleItem, java.io.Serializable, KlavaMonitorable {

  protected Vector tuples ;
  protected NetNode node ;
  protected KlavaMonitorable monitorable ;

  public TupleSpace() {
    tuples = new Vector() ;
    monitorable = newKlavaMonitorable() ; // factory method
  }
  public TupleSpace( Vector v ) { tuples = v ; }
  public TupleSpace( TupleSpace ts ) { tuples = (Vector)ts.tuples.clone() ; }

  // shows the TS graphically: to be implemented by derived classes
  public void show() {}

  public void setNode( NetNode n ) { node = n ; }

  protected KlavaMonitorable newKlavaMonitorable() { return new KlavaMonitorableDefault() ; }

  // KlavaMonitorable interface
  public void addMonitor( KlavaMonitor m ) { monitorable.addMonitor( m ) ; }
  public void removeMonitor( KlavaMonitor m ) { monitorable.removeMonitor( m ) ; }
  public void notifyAddElement( Object o ) { monitorable.notifyAddElement( o ) ; }
  public void notifyRemoveElement( Object o ) { monitorable.notifyRemoveElement( o ) ; }
  public void notifyRemoveAll() { monitorable.notifyRemoveAll() ; }

  public void out( Tuple t ) {
    _out( t ) ;
  }

  protected void _out( Tuple t ) {
    synchronized ( tuples ) {
      tuples.addElement( t ) ;
      tuples.notifyAll() ; // wakes up all waiting processes
      notifyAddElement (t);
    }
  }

  public void add( Tuple t ) {
      tuples.addElement( t ) ;
  }

  public void outString( String s )
          throws KlavaTupleParsingException
  {
    out( TupleParser.parseString( s ) ) ;
  }

  public boolean read( Tuple t ) {
    return ReadIn( t, false, true ) ;
  }

  public boolean in( Tuple t ) {
    return ReadIn( t, true, true ) ;
  }

  // time out versions
  public boolean read( Tuple t, long TimeOut ) throws KlavaTimeOutException {
    if ( TimeOut < 0 )
      return read( t ) ;
    else
      return ReadIn( t, false, TimeOut ) ;
  }

  public boolean in( Tuple t, long TimeOut ) throws KlavaTimeOutException {
    if ( TimeOut < 0 )
      return in( t ) ;
    else
      return ReadIn( t, true, TimeOut ) ;
  }

    // Timeout boolean versions
    public boolean read_t(Tuple t, long TimeOut)
    {
	try {
	    return read(t, TimeOut);
	} catch (KlavaTimeOutException e) {
	    return false;
	}
    }

    public boolean in_t(Tuple t, long TimeOut)
    {
	try {
	    return in(t, TimeOut);
	} catch (KlavaTimeOutException e) {
	    return false;
	}
    }

  public boolean read_nb( Tuple t ) {
    return ReadIn( t, false, false ) ;
  }

  public boolean in_nb( Tuple t ) {
    return ReadIn( t, true, false ) ;
  }

  protected boolean ReadIn( Tuple t, boolean removeTuple, boolean blocking ) {
    Tuple toMatch ;
    boolean matched = false ;

    synchronized ( tuples ) {
      while ( ! matched ) {
      	for( int i = 0 ; i < length() ; i++ ) {
      	  toMatch = getTuple( i ) ;
      	  if( toMatch.match( t ) ) {
      	    if( removeTuple )
      	      removeTuple( i ) ;
      	    matched = true ;
      	    break ;
      	  }
      	}
      	if ( ! matched ) {
      	  if ( blocking ) {
      	    try {
      	      tuples.wait() ;
      	    } catch ( InterruptedException e ) {
              e.printStackTrace ();
              return false ;
            }
      	  }
      	  else
      	    return false ;
      	}
      } // while ( ! matched )
    } // synchronized ( tuples )

    // if we're here we found a matching tuple
    return true ;
  }

  // time out version
  protected boolean ReadIn( Tuple t, boolean removeTuple, long TimeOut )
    throws KlavaTimeOutException {
      Tuple toMatch ;
      boolean matched = false ;

      synchronized ( tuples ) {
      	long waitTime = TimeOut ;
      	long startTime = System.currentTimeMillis() ;
      	long timeSoFar ;
      	while ( ! matched ) {
      	  for( int i = 0 ; i < length() ; i++ ) {
      	    toMatch = getTuple( i ) ;
      	    if( toMatch.match( t ) ) {
      	      if( removeTuple )
           		removeTuple( i ) ;
       	      matched = true ;
       	      break ;
       	    }
       	  }
      	  if ( ! matched ) {
      	    timeSoFar = System.currentTimeMillis() - startTime ;
      	    if ( timeSoFar >= TimeOut ) {
      	      // TIME OUT !
	            throw new KlavaTimeOutException( "tuple : " + t ) ;
      	    }
      	    try {
      	      waitTime = TimeOut - timeSoFar ;
      	      tuples.wait( waitTime ) ;
      	    } catch ( InterruptedException e ) {
      	      e.printStackTrace();
      	      return false ;
            }
      	    timeSoFar = System.currentTimeMillis() - startTime ;
      	    if ( timeSoFar >= TimeOut ) {
	            // TIME OUT !
      	      throw new KlavaTimeOutException( "tuple : " + t ) ;
      	    }
      	  }
      	} // while ( ! matched )
      } // synchronized ( tuples )

      return true ;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer( "{ " ) ;

    synchronized( tuples ) {
      for ( int i = 0 ; i < length() ; i++ ) {
        sb.append( getTuple( i ) ) ;
        if ( i < length() - 1 ) {
        	sb.append( ", " ) ;
        }
      }
    }
    sb.append( " }" ) ;
    return sb.toString() ;
  }

  public int length() {
    return tuples.size() ;
  }

  protected Tuple getTuple( int index ) {
    return (Tuple)tuples.elementAt( index ) ;
  }

  protected void removeTuple( int i ) {
    synchronized ( tuples )
    {
      Object o = tuples.elementAt (i);
      tuples.removeElementAt (i);
      notifyRemoveElement (o);
    }
  }

  protected void removeAllTuples() {
    synchronized (tuples) {
      tuples.removeAllElements();
      notifyRemoveAll();
    }
  }

  public Vector getTuples() { return tuples ; }

  // TupleItem interface
  public boolean isFormal() { return length() == 0 ; }
  public void setFormal() { tuples = new Vector() ; }
  public void setValue( Object o ) {
    if ( o != null && o instanceof TupleSpace )
      tuples = ((TupleSpace)o).getTuples() ;
  }
  public boolean equals( Object o ) {
    if ( o != null && o instanceof TupleSpace )
      return true ;
    return false ;
  }
  public Object duplicate() { return new TupleSpace(this) ; }
}
