package Klava.crypto;

import Klava.TupleItem;

import java.io.*;

import java.security.*;

import javax.crypto.*;


/**
 * Represent an encrypted tuple item
 *
 * @author $author$
 * @version $Revision: 1.1.2.1 $
 */
public class KCipher implements TupleItem {
    /** The tuple item value encrypted */
    protected byte[] encItem;

    /**
     * The reference to the real clear-text tuple item (only when this has been
     * successfully decrypted)
     */
    protected Object ref;

    /** Is it a formal tuple item? */
    public boolean formal;

    /** The algorithm used for encryption/decryption */
    public String alg;

    /**
     * Creates a new KCipher object.
     */
    private KCipher() {
    }

    /**
     * Creates a new KCipher object.
     *
     * @param o the clear-text tuple item.  It can be either a formal or an
     *        actual tuple item.
     */
    public KCipher(TupleItem o) {
        encItem = null;

        if (o.isFormal()) {
            ref = o;
            formal = true;
        } else {
            ref = o; // needed for the encryption
            formal = false;
        }

        alg = null;
    }

    /**
     * Creates a new KCipher object.  Copy constructor
     *
     * @param c the KCipher to copy from
     */
    public KCipher(KCipher c) {
        encItem = c.encItem;
        ref = c.ref;
        formal = c.formal;
        alg = c.alg;
    }

    /**
     * Return a string representation
     *
     * @return the string representation of this tuple item.
     */
    public String toString() {
        if (!formal) {
            if (encItem != null) {
                return encItem.toString();
            } else {
                return ((ref != null) ? ref.toString() : "null Object!");
            }
        } else {
            return "!KCipher (" +
            ((ref != null) ? ref.toString() : "null Object!") + ")";
        }
    }

    /**
     * Is this tuple item a formal tuple item
     *
     * @return true if this is a formal tuple item
     */
    public boolean isFormal() {
        return (formal);
    }

    /**
     * set the value of the tuple item
     *
     * @param o the value to set
     */
    public void setValue(Object o) {
        if (!(o instanceof KCipher)) {
            System.err.println("ERROR: setValue called with non KCipher!");

            return;
        }

        KCipher copy = (KCipher) o;

        if (!(copy.ref instanceof TupleItem) && !(copy.ref == null)) {
            System.err.println(
                "ERROR: setValue called with KCipher.ref non TupleItem!");

            return;
        }

        encItem = copy.encItem;
        ((TupleItem) ref).setValue((TupleItem) copy.ref);
        formal = copy.formal;
        alg = copy.alg;
    }

    /**
     * Test equality with another tuple item
     *
     * @param o the other tuple item
     *
     * @return whether this tuple item is equal to the one passed
     */
    public boolean equals(Object o) {
        return false;
    }

    /**
     * Create a duplicate of this tuple item
     *
     * @return a duplicate of this tuple item
     */
    public Object duplicate() {
        KCipher dup = new KCipher(this);
        dup.ref = ((ref instanceof TupleItem) ? ((TupleItem) ref).duplicate()
                                              : ref);

        return dup;
    }

    /**
     * Perform the encryption of this tuple item
     *
     * @param key the key to use for the encryption
     * @param encType the type of encryption
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    protected void enc(Key key, String encType)
        throws IOException, GeneralSecurityException {
        Cipher desCipher;
        desCipher = Cipher.getInstance(encType);
        desCipher.init(Cipher.ENCRYPT_MODE, key);

        ByteArrayOutputStream oStream = new ByteArrayOutputStream();
        ObjectOutputStream out = new ObjectOutputStream(oStream);
        out.writeObject(ref);
        out.flush();
        encItem = desCipher.doFinal(oStream.toByteArray());

        out.close();

        ref = null;
        formal = false;
        alg = encType;
    }

    /**
     * perform the decryption of this tuple item
     *
     * @param key the key to use for decryption
     *
     * @return an array of bytes representing the decrypted data
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    protected byte[] dec(Key key) throws IOException, GeneralSecurityException {
        Cipher desCipher;
        desCipher = Cipher.getInstance(alg);
        desCipher.init(Cipher.DECRYPT_MODE, key);

        byte[] decoded = desCipher.doFinal(encItem);

        return decoded;
    }
}
