package Klava.crypto;

import java.io.*;

import java.security.*;

import java.util.*;

import javax.crypto.*;
import javax.crypto.spec.*;


/**
 * mplement a keystore. Upon creation the passphrase for the requested keystore
 * is required.
 *
 * @author $author$
 * @version $Revision: 1.1.2.2 $
 */
public class KStore {
    /**
     * Code for the KStoreFileAccess.
     *
     * @see KStoreFileAccess
     */
    public static final int FILE = 0;

    /**
     * Code for the KStoreDialogAccess.
     *
     * @see KStoreDialogAccess
     */
    public static final int DIALOG = 1;

    /**
     * Code for the KStoreTextAccess.
     *
     * @see KStoreTextAccess
     */

    public static final int CONSOLE = 2;

    /** The keystore */
    private KeyStore kKeyStore = null;

    /** the reference to the access method to the keystore
     * used for requesting a password and the key of the keystore */
    public KStoreAccess kstore_access;

    /**
     * Default constructor. The keystore is set to null and the KStoreAccess is
     * set to KStoreDialogAccess.
     *
     * @see KStoreDialogAccess
     */
    public KStore() {
        kstore_access = new KStoreDialogAccess();
    }

    /**
     * Constructor. Set the specified KStoreAccess
     *
     * @param kstore_accessType the type of KStoreAccess to be used, it can be
     *        KStore.FILE, KStore.DIALOG, KStore.CONSOLE (no default)
     * @param file_name only makes sense if a KStore.FILE is requested, since
     *        it represents the name of that file
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public KStore(int kstore_accessType, String file_name)
        throws IOException, GeneralSecurityException {
        if (kstore_accessType == KStore.FILE) {
            kstore_access = new KStoreFileAccess(file_name);
        } else if (kstore_accessType == KStore.DIALOG) {
            kstore_access = null; //new KStoreDialogAccess() ;
        } else if (kstore_accessType == KStore.CONSOLE) {
            kstore_access = new KStoreTextAccess();
        } else {
            throw new GeneralSecurityException("Invlaid keystore_accessType");
        }
    }

    /**
     * Constructor. Open the specified keystore with the specified password
     *
     * @param storeName the name of the file with the keystore
     * @param storePassword the password associated with the keystore
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public KStore(String storeName, String storePassword)
        throws IOException, GeneralSecurityException {
        kstore_access = new KStoreTextAccess();
        openKeyStore(storeName, storePassword);
    }

    /**
     * Constructor. Use the specified KStoreAccess.
     *
     * @param access the KStoreAccess to be used.
     */
    public KStore(KStoreAccess access) {
        kstore_access = access;
    }

    /**
     * Set the object for accessing the keystore
     *
     * @param access the object for accessing the keystore
     */
    public void setKStoreAccess(KStoreAccess access) {
        kstore_access = access;
    }

    /**
     * Open the keystore after the passphrase is given
     *
     * @param storeName the name of the file with the keystore
     * @param storePass the password associated with the keystore
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void openKeyStore(String storeName, String storePass)
        throws IOException, GeneralSecurityException {
        kKeyStore = KeyStore.getInstance("JCEKS");

        FileInputStream storeis = new FileInputStream(storeName);
        char[] pass = storePass.toCharArray();
        kKeyStore.load(storeis, pass);
    }

    /**
     * Is the keystore already open?
     *
     * @return true if the keystore is open, false otherwise
     */
    public boolean isKeyStoreOpen() {
        return (kKeyStore != null);
    }

    /**
     * return the size of the keystore
     *
     * @return the size of the keystore
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public int size() throws IOException, GeneralSecurityException {
        return kKeyStore.size();
    }

    /**
     * Return the list of the keys in the keystore.
     *
     * @return the Enumeration of the keys in the keystore
     *
     * @throws IOException
     * @throws GeneralSecurityException if the keystore is not open
     */
    public Enumeration getKeyAliases()
        throws IOException, GeneralSecurityException {
        if (kKeyStore == null) {
            throw new GeneralSecurityException("The keystore is not open");
        }

        return kKeyStore.aliases();
    }

    /**
     * Return a specific Key
     *
     * @param alias the alias for the key
     * @param password the password for the key
     *
     * @return the Key
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    private Key getKey(String alias, String password)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        return kKeyStore.getKey(alias, password.toCharArray());
    }

    /**
     * Return a specific Key, using a specific message
     *
     * @param message the message to request the key
     *
     * @return the Key
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public synchronized Key getKey(String message)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        KeyInfo key_info = kstore_access.askKey(message, this);
        String key_alias = key_info.getKeyName();
        String key_pass = key_info.getKeyPassword();

        if ((key_alias == null) || (key_alias.length() == 0)) {
            throw new GeneralSecurityException("Unspecified Key Alias");
        }

        if ((key_pass == null) || (key_pass.length() == 0)) {
            throw new GeneralSecurityException("Unspecified Key Password");
        }

        return getKey(key_alias, key_pass);
    }

    /**
     * Import a key into the keystore
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void importKey()
        throws FileNotFoundException, IOException, GeneralSecurityException {
        /* import encoded key from a file */
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String keyFileName;
        System.out.println(
            ">ks>specifiy the file with the key to be imported");
        keyFileName = in.readLine();

        FileInputStream keyfis = new FileInputStream(keyFileName);
        byte[] encKey = new byte[keyfis.available()];
        keyfis.read(encKey);
        keyfis.close();

        DESKeySpec desKeySpec = new DESKeySpec(encKey);
        SecretKeyFactory keyFactory = SecretKeyFactory.getInstance("DES",
                "SunJCE");
        SecretKey desKey = keyFactory.generateSecret(desKeySpec);
        this.addKey(desKey);
    }

    /**
     * Generate a key and add it to this keystore
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void genKey()
        throws FileNotFoundException, IOException, GeneralSecurityException {
        /* Generate a key  */
        System.out.println(">ks>Creating the key ........");

        KeyGenerator keygen = KeyGenerator.getInstance("DES", "SunJCE");
        SecretKey desKey = keygen.generateKey();
        this.addKey(desKey);
    }

    /**
     * Add a key to this keystore
     *
     * @param desKey the secret key
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void addKey(SecretKey desKey)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String keyAlias;
        System.out.println(">ks>Name for this key?");
        keyAlias = in.readLine();

        String keyPass;
        System.out.println(">ks>passphrase for this key?");
        keyPass = in.readLine();
        kKeyStore.setKeyEntry(keyAlias, desKey, keyPass.toCharArray(), null);
    }

    /**
     * Delete a key from this keystore
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void delKey()
        throws FileNotFoundException, IOException, GeneralSecurityException {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String keyAlias;
        System.out.println(">ks>Which key do you want to remove?");
        keyAlias = in.readLine();
        kKeyStore.deleteEntry(keyAlias);
    }

    /**
     * Save the keystore into a file
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public void save()
        throws FileNotFoundException, IOException, GeneralSecurityException {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String storeName;
        System.out.println(">ks>Name of the file where to save the keystore?");
        storeName = in.readLine();

        String storePass;
        System.out.println(">ks>Passphrase for the keystore?");
        storePass = in.readLine();

        FileOutputStream storeos = new FileOutputStream(storeName);
        kKeyStore.store(storeos, storePass.toCharArray());
        storeos.close();
    }
}
