package Klava.crypto;

import Klava.KlavaException;
import Klava.KlavaProcess;
import Klava.Locality;
import Klava.NodeMessage;

import java.io.FileNotFoundException;
import java.io.IOException;

import java.security.GeneralSecurityException;
import java.security.Key;


/**
 * Specialize KlavaProcess with a personal KeyStore
 *
 * @author $author$
 * @version $Revision: 1.1.2.1 $
 */
public abstract class KlavaProcessx extends KlavaProcess {
    /**
     * The keystore of the process. As it is transient it is not delivered if
     * the process is sent to a remote computer, or it is serialized.
     */
    transient protected KStore myKeyStore;

    /** The key used for encryption */
    transient private Key encKey = null;

    /** The key used for decryption */
    transient private Key decKey = null;

    /**
     * Constructor. Initialize the keystore of the process with a new keystore.
     */
    public KlavaProcessx() {
        myKeyStore = new KStore();
    }

    /**
     * Constructor. Use the specified keystore
     *
     * @param storeName the name of the file with the key store
     * @param storePass the password associated with the keystore
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public KlavaProcessx(String storeName, String storePass)
        throws IOException, GeneralSecurityException {
        myKeyStore = new KStore(storeName, storePass);
    }

    /**
     * Constructor. Use the specified keystore
     *
     * @param access the KStoreAccess that is used to request info about keys
     *        and password
     *
     * @throws IOException
     * @throws GeneralSecurityException
     */
    public KlavaProcessx(KStoreAccess access)
        throws IOException, GeneralSecurityException {
        myKeyStore = new KStore(access);
    }

    /**
     * Get the keystore of the process. This method is obviously accessible
     * only from within the package
     *
     * @return the keystore of the process.
     */
    KStore getKStore() {
        return myKeyStore;
    }

    /**
     * Set the key that will always used (until unset) for encryption
     *
     * @param message the message for requesting the key
     *
     * @throws FileNotFoundException the key file is not found
     * @throws IOException
     * @throws GeneralSecurityException
     */
    protected void setEncKey(String message)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        encKey = myKeyStore.getKey(message);
    }

    /**
     * Set the key that will always used (until unset) for decryption
     *
     * @param message the message for requesting the key
     *
     * @throws FileNotFoundException the key file is not found
     * @throws IOException
     * @throws GeneralSecurityException
     */
    protected void setDecKey(String message)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        decKey = myKeyStore.getKey(message);
    }

    /**
     * unset the default encryption key
     */
    protected void unsetEncKey() {
        encKey = null;
    }

    /**
     * unset the default decryption key
     */
    protected void unsetDecKey() {
        decKey = null;
    }

    /**
     * If no default encryption key is set, ask the user to choose one
     *
     * @param message the message used to ask for the key
     *
     * @return the chosen key
     *
     * @throws FileNotFoundException the key file is not found
     * @throws IOException
     * @throws GeneralSecurityException
     */
    protected Key useEncKey(String message)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        if (encKey != null) {
            return encKey;
        } else if (myKeyStore != null) {
            Key key = myKeyStore.getKey(message);

            return key;
        } else {
            throw new GeneralSecurityException("Missing KeyStore ");
        }
    }

    /**
     * If no default decryption key is set, ask the user to choose one
     *
     * @param message the message used to ask for the key
     *
     * @return the chosen key
     *
     * @throws FileNotFoundException the key file is not found
     * @throws IOException DOCUMENT ME!
     * @throws GeneralSecurityException DOCUMENT ME!
     */
    protected Key useDecKey(String message)
        throws FileNotFoundException, IOException, GeneralSecurityException {
        if (decKey != null) {
            return decKey;
        } else if (myKeyStore != null) {
            Key key = myKeyStore.getKey(message);

            return key;
        } else {
            throw new GeneralSecurityException("Missing KeyStore ");
        }
    }

    /**
     * Automatic <b>in</b> and decryption: it searches for a matching tuple
     * with encrypted fields, when it finds one it removes the matching tuple
     * and it tries decryption if the decryption fails put the tuple back and
     * searches for another one.  If no new matching tuple is found, it blocks
     * the process until a new matching tuple is available.
     *
     * @param t the template tuple
     * @param l the destination locality
     *
     * @throws KlavaException
     */
    final protected void ink(Tuplex t, Locality l) throws KlavaException {
        readink(t, l, -1, NodeMessage.IN, true);
    }

    /**
     * Non blocking version of ink: if the decryption fails it still tries with
     * another matching tuple; if no new tuple is found it simply returns
     * <tt>false</tt>
     *
     * @param t the template tuple
     * @param l the destination locality
     *
     * @return true if a tuple is successfully retrieved and decrypted
     *
     * @throws KlavaException
     */
    final protected boolean ink_nb(Tuplex t, Locality l)
        throws KlavaException {
        return readink(t, l, -1, NodeMessage.IN, false);
    }

    /**
     * Timeout version of ink
     *
     * @param t the template tuple
     * @param l the destination locality
     * @param TimeOut
     *
     * @throws KlavaException
     */
    final protected void ink(Tuplex t, Locality l, long TimeOut)
        throws KlavaException {
        readink(t, l, TimeOut, NodeMessage.IN, true);
    }

    /**
     * Automatic <b>read</b> and decryption: it searches for a matching tuple
     * with encrypted fields, when it finds one it tries decryption if the
     * decryption fails it searches for another one.  If no new matching tuple
     * is found, it blocks the process until a new matching tuple is
     * available.
     *
     * @param t the template tuple
     * @param l the destination locality
     *
     * @throws KlavaException DOCUMENT ME!
     */
    final protected void readk(Tuplex t, Locality l) throws KlavaException {
        readink(t, l, -1, NodeMessage.READ, true);
    }

    /**
     * Non blocking version of readk: if the decryption fails it still tries
     * with another matching tuple; if no new tuple is found it simply returns
     * <tt>false</tt>
     *
     * @param t the template tuple
     * @param l the destination locality
     *
     * @return true if a tuple is successfully retrieved and decrypted
     *
     * @throws KlavaException
     */
    final protected boolean readk_nb(Tuplex t, Locality l)
        throws KlavaException {
        return readink(t, l, -1, NodeMessage.READ, false);
    }

    /**
     * Timeout version of readk
     *
     * @param t the template tuple
     * @param l the destination locality
     * @param TimeOut
     *
     * @throws KlavaException
     */
    final protected void readk(Tuplex t, Locality l, long TimeOut)
        throws KlavaException {
        readink(t, l, TimeOut, NodeMessage.READ, true);
    }

    /**
     * Internal method actually performing all the issues related to an ink and
     * readk.
     *
     * @param t the template tuple
     * @param l the destination locality
     * @param Timeout
     * @param opcode the actual opcode of the operation
     * @param blocking whether it is a blocking action
     *
     * @return true if the action succeeds
     *
     * @throws KlavaException
     * @throws Klava.KlavaTimeOutException
     */
    protected boolean readink(Tuplex t, Locality l, long Timeout, int opcode,
        boolean blocking) throws KlavaException {
        KlavaProcessx requester = (KlavaProcessx) Thread.currentThread();
        boolean done = false;
        boolean result = false;
        long startTime = System.currentTimeMillis();
        long elapsedTime;
        boolean requested_timeout = (Timeout >= 0);

        while (!done) {
            try {
                result = super.readin(t, l, opcode, Timeout, blocking);

                if (result) {
                    if (t.matchTuplex(requester)) {
                        done = true; // this will be checked in the "finally"

                        // that will be executed anyway
                        return true;
                    }
                } else if (!blocking) {
                    done = true; // this will be checked in the "finally"

                    // that will be executed anyway
                    return false;
                }
            } catch (java.security.InvalidKeyException e1) {
                //e1.printStackTrace ();
                System.err.println(
                    "The key doesn't exist or it's for a different algotithm");
            } catch (java.security.UnrecoverableKeyException e2) {
                //e2.printStackTrace ();
                System.err.println(
                    "The password is wrong or the key has been compromised");
            } catch (javax.crypto.BadPaddingException e3) {
                //e3.printStackTrace ();
                System.err.println("You have specified a wrong key");
            } catch (Klava.KlavaTimeOutException te) {
                throw te; // otherwise it is caught by the next catch
            } catch (Klava.KlavaException ke) {
                throw ke;
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                if (!done) {
                    // otherwise we never end
                    // if we are here it's because decryption failed
                    if (opcode == NodeMessage.IN) {
                        out(t.getMatched(), l);
                    }

                    t.resetOriginalTemplate();
                }
            }

            if (requested_timeout) {
                if (Timeout >= 0) { // recompute timeout
                    elapsedTime = System.currentTimeMillis() - startTime;
                    Timeout -= elapsedTime;

                    if (Timeout <= 0) {
                        throw new Klava.KlavaTimeOutException("tuple : " + t);
                    }
                } else {
                    throw new Klava.KlavaTimeOutException("tuple : " + t);
                }
            }
        }

        return false;
    }
}
