package Klava.debug;

import java.util.List ;
import java.util.ArrayList ;
import java.util.Collections ;

import Klava.NodeInterface ;
import Klava.KlavaProcess ;
import Klava.Tuple ;
import Klava.Locality ;
import Klava.PhysicalLocality ;
import Klava.KlavaException ;
import Klava.Queue ;

import Klava.debug.event.* ;

public class KlavaAgent {

  private String id ;
  private KlavaProcessInstance currentProcess ;
  private NodeInterface currentNode ;

  // Status constants
  public static final int RUNNING = 0 ;
  public static final int SUSPENDED = 1 ;
  public static final int MIGRATING = 2 ;
  public static final int TERMINATED = 3 ;

  private int status = RUNNING ;

  // To handle breakpoints
  private ArrayList breakpoints = new ArrayList() ;
  private List breakpointsClone = Collections.EMPTY_LIST ;

  // To implement Kill
  private boolean killed = false ;
  private boolean atBlockingAction = false ;

  // To listen for processes' step events
  private StepHandler stepHandler ;
  // To store generated events
  private Queue eventQueue ;

  KlavaAgent( String id, KlavaProcess currentP,
              NodeInterface currentNode, Queue eventQueue )
  {
    this.id = id ;
    this.currentNode = currentNode ;
    this.eventQueue = eventQueue ;
    this.stepHandler = new StepHandler() ;
    buildProcCallStack( currentP ) ;
  }

  private void buildProcCallStack( KlavaProcess P ) {
    KlavaProcess migrCaller = P.getMigratableCaller() ;
    if ( migrCaller != null ) {
      // Recursive call
      buildProcCallStack( migrCaller ) ;
    }
    pushProcess( P ) ;
  }

  private void pushProcess( KlavaProcess P ) {
    if ( P instanceof KlavaDebugProcess ) {
      KlavaDebugProcess dbP = ( KlavaDebugProcess ) P ;
      dbP.setAttached( true ) ;
      dbP.setStepListener( stepHandler ) ;
    }
    currentProcess = new KlavaProcessInstance( P, currentNode, currentProcess ) ;
  }

  private KlavaProcess popProcess() {
    KlavaProcess poppedProcess = currentProcess.getProc() ;
    currentProcess = currentProcess.getCaller() ;
    if ( poppedProcess instanceof KlavaDebugProcess ) {
      KlavaDebugProcess dbP = ( KlavaDebugProcess ) poppedProcess ;
      dbP.setAttached( false ) ;
      dbP.setStepListener( null ) ;
    }
    return poppedProcess ;
  }

  public String getId() {
    return id ;
  }

  public PhysicalLocality getLocality() {
    return currentNode.ThisLocality() ;
  }

  public synchronized int getStatus() {
    return status ;
  }

  public synchronized void resume() {
    if ( status == SUSPENDED ) {
      currentProcess.setValid( false ) ;
      Print( "Resuming " + id ) ;
      status = RUNNING ;
      notify() ;
    }
  }

  public synchronized void kill() {
    if ( ! killed  ) {
      killed = true ;
      Print( "Killing " + id ) ;
      if ( atBlockingAction ) {
        currentProcess.getProc().interrupt() ;
      }
      resume() ;
    }
  }

  public void addBreakpoint( Breakpoint b ) {
    if ( b == null )
      return ;
    synchronized( breakpoints ) {
      breakpoints.add( b ) ;
      breakpointsClone = ( List ) breakpoints.clone() ; // Atomic assign
    }
  }

  public void removeBreakpoint( int bid ) {
    synchronized( breakpoints ) {
      int index = -1 ;
      int bCount = breakpoints.size() ;
      for ( int i = 0 ; i < bCount ; i++ ) {
        if ( ( ( Breakpoint ) breakpoints.get( i ) ).getId() == bid )
          break ;
      }
      if ( index != -1 ) { // Found
        breakpoints.remove( index ) ;
        breakpointsClone = ( List ) breakpoints.clone() ; // Atomic assign
      }
    }
  }

  public Breakpoint[] getBreakpoints() {
    return ( Breakpoint[] ) breakpointsClone.toArray() ;
  }

  public synchronized KlavaProcessInstance getCurrentProcess()
    throws AgentNotSuspendedException
  {
    if ( isCurrentProcAccess() )
        return currentProcess ;

    if ( status == SUSPENDED ) {
      currentProcess.setValid( true ) ;
      return currentProcess ;
    } else {
      throw new AgentNotSuspendedException( id ) ;
    }
  }

  private boolean isCurrentProcAccess() {
    if ( currentProcess == null ) { // agent terminated
      return false ;
    } else {
      return ( Thread.currentThread() == currentProcess.getProc() ) ;
    }
  }

  // Sanity check
   private void checkCurrentProc() {
     if ( ! isCurrentProcAccess() )
       throw new IllegalStateException
       ( "Not current process access!" ) ;
  }

  private void Print( String s ) {
    currentNode.Print( "[DEBUG]->" + s ) ;
  }

  private void locationPrint( String s ) {
    Print( s + " at " + currentProcess.getLocation() ) ;
  }

  private void sendEvent( AgentEvent ae ) {
    eventQueue.add( ae ) ;
  }

  // Package-friendly interface
  // These methods are called by DebugNode in the current process
  // thread of control to mantain a causal connection between
  // running agents and their meta-objects counterpart

  void onTSActionEntry( TSAction.Op op, Tuple t, Locality l, boolean blocking ) {
    checkCurrentProc() ;
    if ( blocking ) {
      synchronized ( this ) {
        atBlockingAction = true ;
        if ( killed ) {
          currentProcess.getProc().interrupt() ;
        }
      }
    }
  }

  void onTSActionExit( TSAction.Op op, Tuple et, PhysicalLocality pl, boolean succeeded ) {
    checkCurrentProc() ;
    synchronized ( this ) {
      if ( atBlockingAction ) {
        atBlockingAction = false ;
      }
    }
    if ( succeeded ) {
      currentProcess.setCurrentAction( new TSAction( op, et, pl ) ) ;
    }
  }

  void onProcCallEntry( KlavaProcess P ) {
    checkCurrentProc() ;
    pushProcess( P ) ;
  }

  void onProcCallExit( KlavaProcess P ) {
    checkCurrentProc() ;
    popProcess() ;
    if ( currentProcess == null ) { // process Stack empty
      status = TERMINATED ;
      sendEvent( new AgentEndEvent( id, getLocality() ) ) ;
    }
  }

  void onException( KlavaException ke ) {
    checkCurrentProc() ;
    if ( status == MIGRATING ) {
      status = RUNNING ;
    }
    sendEvent( new ExceptionEvent( id, getLocality(), currentProcess.getLocation(), ke ) ) ;
  }

  void onMigration( PhysicalLocality destLoc ) {
    checkCurrentProc() ;
    if ( isAllStackMigrating() ) {
      status = MIGRATING ;
      sendEvent( new AgentMigrationEvent( id, getLocality(), currentProcess.getLocation(), destLoc ) ) ;
    }
  }

  /* Due to security considerations migrations performed by processes retrieved
     from some tuple space dont have effect on entire stack of called processes */
  private boolean isAllStackMigrating() {
    KlavaProcess currentP = currentProcess.getProc() ;// unwrap current proc
    // Walk until we have a caller that can migrates
    while( ( currentP.getMigratableCaller() ) != null  ) {
      currentP = currentP.getMigratableCaller() ;
    }
    // This means that we have only procs that migrates w/ the current proc
    return ( currentP.getCaller() == null ) ;
  }

  // StepListener impl
  private class StepHandler implements KlavaDebugProcess.StepListener {
    public void processStepped( KlavaDebugProcess P ) throws KillException {
      // If killed unwind process stack by sending an interrupt
      if ( killed ) {
        throw new KillException( "Forced termination" ) ;
      }

      // Check for breakpoint events
      // this can trigger suspension
      checkBreakpoints() ;

      // Clear performed action for next stopping point
      currentProcess.setCurrentAction( null ) ;
    }
  }

  // Check breakpoints
  private void checkBreakpoints() {

    List bkpts = breakpointsClone ;
    int bkptsCount = bkpts.size() ;

    Breakpoint b ;
    BreakpointEvent bEvent = null ;
    BreakpointEvents bEvents = null ; // multiple hits

    for ( int i = 0 ; i < bkptsCount ;  i++ ) {
      b = ( Breakpoint ) bkpts.get( i ) ;
      try {
        bEvent = b.locationChanged( this ) ;
      } catch ( Exception e ) { // Should never happen
        e.printStackTrace() ;
        Print( "Exception in " + b + " : " + e ) ;
        Print ( "Remove it and continue..." ) ;
        removeBreakpoint( b.getId() ) ;
      }

      if ( bEvent != null ) { // it's been hit
        if ( bEvents == null ) {
          bEvents = new BreakpointEvents( bEvent ) ;
        } else {
          bEvents.addEvent( bEvent ) ;
        }
      }

    } // end loop

    if ( bEvents != null ) { // At least one hit
      handleBreakpointEvents( bEvents ) ;
    }
  }

  private synchronized void handleBreakpointEvents( BreakpointEvents bEvents ) {
    locationPrint( "Hit breakpoint(s)" ) ;
    sendEvent( bEvents ) ;
    if ( (bEvents.triggerSuspension()) && ( !killed ) ) {
      status = SUSPENDED ;
      while ( ( status == SUSPENDED)  &&   ( !killed )  )  {
        try {
          wait() ;
        } catch ( InterruptedException ie ) {
          System.err.println( ie ) ;
        }
      }
    }
  }

}
