// Mirrors KlavaProcess instance in agent call stack
package Klava.debug ;

import java.lang.reflect.Field ;
import java.lang.reflect.AccessibleObject ;
import java.util.Map ;
import java.util.HashMap ;
import java.util.Iterator ;
import Klava.* ;

public class KlavaProcessInstance {

  private KlavaProcess P ;
  private NodeInterface node ; // to eval logical localities

  private Location location ;
  private KlavaProcessInstance caller ;
  private int callDepth ;
  private Map varFields = null ; // name -> field
  private boolean isValid = false ;
  private TSAction currentAction = null ;

  KlavaProcessInstance ( KlavaProcess P, NodeInterface node, KlavaProcessInstance caller )
  {
    this.P = P ;
    this.node = node ;
    this.caller = caller ;
    location = new Location( P.KlavaProcName(), -1, -1 ) ;
    callDepth = ( caller == null ? 1 : caller.callDepth + 1 ) ;
  }

  public KlavaProcessInstance( KlavaProcess P, NodeInterface node ) {
    this( P, node, null ) ;
  }

  // Accessible only in this package
  KlavaProcess getProc() {
    return P ;
  }

  void setValid( boolean valid ) {
    isValid = valid ;
    if ( caller != null ) { // Recursive call to caller
      caller.setValid( valid ) ;
    }
  }

  private void checkValid() throws InvalidProcessInstanceException {
    if ( Thread.currentThread() != P  &&  ! isValid  ) {
      throw new InvalidProcessInstanceException() ;
    }
  }

  public Location getLocation() {
    checkValid() ;
    refreshLocation() ;
    return location ;
  }

  private void refreshLocation() {
    if ( P instanceof KlavaDebugProcess ) {
      KlavaDebugProcess dbP = ( KlavaDebugProcess ) P ;
      int currLine = dbP.getCurrentLine() ;
      int currCol = dbP.getCurrentColumn() ;
      if ( location.line != currLine || location.column != currCol ) {
        location = new Location( dbP.KlavaProcName(), currLine, currCol ) ;
      }
    }
  }

  public Object getValue( String varName ) {
    checkValid() ;
    Field f = ( Field ) getVarFields().get( varName ) ;
    if ( f != null ) {
      return getValue( f ) ;
    } else {
      throw new IllegalArgumentException
      (  varName + " not found in " + P.KlavaProcName() ) ;
    }
  }

  public Map getAllValues() {
    checkValid() ;
    Map varFields = getVarFields() ;
    Map values = new HashMap( varFields.size() ) ;
    Iterator it = varFields.values().iterator() ;
    Field f ;
    while( it.hasNext() ) {
      f = ( Field ) it.next() ;
      values.put( f.getName(), getValue( f ) ) ;
    }
    return values ;
  }

  protected Map getVarFields() {
    if ( varFields == null )  { // Lazy creation

      // Get the process declared fields and make them accessible
      Field[] declFields = P.getClass().getDeclaredFields() ;
      AccessibleObject.setAccessible( declFields, true ) ;

      // Get the self field from superclass and make it accessible
      String self = "self" ;
      Field selfField = null ;
      try {
        selfField = KlavaProcess.class.getDeclaredField( self ) ;
        selfField.setAccessible( true ) ;
      } catch ( NoSuchFieldException nsfe ) { // Can't happen
        nsfe.printStackTrace() ;
      }

      int fieldCount = declFields.length ;
      varFields = new HashMap( fieldCount + 1 ) ;

      // Add self
      varFields.put( self, selfField ) ;

      // Add all other declared fields for variables
      // skipping dummies for used processes
      Field declField ;
      for ( int i = 0 ; i < fieldCount ; i++ ) {
        declField = declFields[ i ] ;
        if ( ! KlavaProcess.class.isAssignableFrom( declField.getType() ) ) {
          varFields.put( declField.getName(), declField ) ;
        }
      }
    }

    return varFields ;
  }

  protected Object getValue( Field field ) {
    Object value = null ;
    try {
      value = field.get( P ) ;
      if ( value instanceof Locality ) {
        value = evaluateLocality( ( Locality ) value ) ;
      }
    } catch ( IllegalAccessException iae ) {
      // Can't happen since all fields are made accessible
      iae.printStackTrace() ;
    }
    return value ;
  }

  protected Locality evaluateLocality( Locality loc ) {
    try {
      if ( loc instanceof PhysicalLocality )
        return loc ;
      PhysicalLocality phyLoc = P.toPhysical( loc ) ;
      if ( phyLoc != null )
        return phyLoc ;
      return node.getPhysicalLocality( loc ) ;
    } catch ( KlavaLogicalLocalityException lle ) {
      /* If loc can't be evaluated w/ available envs act as no op  */
      return loc ;
    }
  }

  public KlavaProcessInstance getCaller() {
    checkValid() ;
    return caller ;
  }

  public int getCallDepth() {
    checkValid() ;
    return callDepth ;
  }

  public TSAction getCurrentAction() {
    return currentAction ;
  }

  void setCurrentAction( TSAction tsAction ) {
    currentAction = tsAction ;
  }

  public String toString() {
    return getLocation().toString() ;
  }

}