/* Meta class that represents klaim actions for accessing TS.
By the use of wildcards( implemented w/ nulls ) we can construct
abstract actions, i.e. not completly specified actions. */
package Klava.debug ;

import java.io.Serializable ;
import java.io.ObjectStreamException ;
import java.io.InvalidObjectException ;
import Klava.PhysicalLocality ;
import Klava.Tuple ;
import Klava.TupleItem ;

public class TSAction implements Serializable {

  // TS action op enum
  public static final class Op implements Serializable {

    public final String name ;

    private Op( String name ) {
      this.name = name ;
    }

    public String toString() {
      return name ;
    }

    public static Op forName( String name )
      throws IllegalArgumentException
    {
      if ( name.equalsIgnoreCase( ANY.name ) )
        return ANY ;
      if ( name.equalsIgnoreCase( OUT.name ) )
        return OUT ;
      if ( name.equalsIgnoreCase( IN.name ) )
        return IN ;
      if ( name.equalsIgnoreCase( READ.name ) )
        return READ ;

      throw new IllegalArgumentException( "Unknown op name" ) ;
    }

    // To avoid duplicates due to deserialization
    private Object readResolve() throws ObjectStreamException {
      try {
        return forName( this.name ) ;
      } catch ( IllegalArgumentException iae ) {
        throw new InvalidObjectException( getClass().getName() ) ;
      }
    }

    public static final Op ANY  = new Op( "any" ) ;
    public static final Op OUT  = new Op( "OUT" ) ;
    public static final Op IN   = new Op( "IN" ) ;
    public static final Op READ = new Op( "READ" ) ;

  }

  private Op op ;
  private Tuple tuple ;
  private PhysicalLocality target ;

  public TSAction( Op op, Tuple tuple, PhysicalLocality target ) {
    this.op = op ;
    this.tuple = tuple ;
    this.target = target ;
  }

  public Op getOp() {
    return op ;
  }

  public Tuple getTuple() {
    return tuple ;
  }

  public PhysicalLocality getTarget() {
    return target ;
  }

  public boolean matches( TSAction otherAction ) {
    return ( otherAction != null &&
             matchesOp( otherAction.op ) &&
             matchesTarget( otherAction.target ) &&
             matchesTuple( otherAction.tuple ) ) ;
  }

  public boolean matchesOp( Op otherOp ) {
    return ( otherOp == Op.ANY || otherOp == op ) ;
  }

  public boolean matchesTuple( Tuple otherTuple ) {
    if ( otherTuple == null )
      return true ;

    if ( tuple == null )
      return false ;

    // To avoid side effects on tuple param
    // due to pattern matching bindings
    Tuple toBeMatched = deepCopy( otherTuple ) ;
    // Delgate to Tuple.match
    return tuple.match( toBeMatched ) ;
  }

  public boolean matchesTarget( PhysicalLocality otherTarget ) {
    return ( otherTarget == null  || otherTarget.equals( target ) ) ;
  }

  // Tuple deep copy
  protected Tuple deepCopy( Tuple t ) {
    // Do first a shallow copy
    Tuple copy = ( Tuple ) t.clone() ;
    Object item ;
    // Clone tuple items
    for ( int i = 0 ; i < copy.length() ; i++ ) {
      item = copy.getItem( i ) ;
      if ( item instanceof TupleItem ) {
        copy.setItem( i, ( ( TupleItem ) item ).duplicate() ) ;
      }
    }
    return copy ;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer( op.toString() ) ;
    sb.append( '(' ) ;
    sb.append( tuple == null  ?  "any"  :  tuple.toString() ) ;
    sb.append( ')' ) ;
    sb.append( '@' ) ;
    sb.append( target == null ?  "any"  :  target.toString() ) ;
    return sb.toString() ;
  }

}