package klavatest;

/**
 * test dispatching of messages in a hierarchical way
 *
 *                       server
 *                    ^           ^
 *                gateway      client
 *                ^
 *          printer_manager
 *           ^
 *       printer
 *
 * printer_client issues a tuple for printers and server has to
 * deliver it to printer, though printer is not directly connected to it.
 * This tests propagation of PROPAGATE_LOCALITY messages to the server.
 */

import Klava.*;

public class testPhysLocReach6 extends KlavaTestApplication {
  public testPhysLocReach6(String name, String suffix,
                                 boolean log_to_file)
                                 throws java.io.IOException,
                                 InterruptedException, KlavaException {
    super(name, suffix, log_to_file);

    NetNode server = new RegisterNetNode("server", 9999);
    server.setNoLogonServer(true); // this way it does not forward messages
    // to any other server

    NetNode printer_manager = new RegisterNetNode("printer_manager", 9998);

    NetNode gateway = new RegisterNetNode("gateway", 9997);

    NetNode printer_client = new NetNode("printer_client", 20001);

    NetNode printer = new NetNode("printer", 10001);

    PhysicalLocality server_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 9999));

    PhysicalLocality printer_manager_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 9998));

    PhysicalLocality printer_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 10001));

    PhysicalLocality gateway_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 9997));

    server.addToEnv("printer", printer_loc.toString());

    addNode(server);
    server.logRoutes("server");

    addNode(gateway);
    gateway.logRoutes("gateway");

    sortFiles("RoutingTable");

    server.start();
    gateway.start();
    printer_manager.start();
    printer.start();
    printer_client.start();

    LogicalLocality printer_log_loc = new LogicalLocality("printer");
    LogicalLocality gateway_log_loc = new LogicalLocality("gateway");
    LogicalLocality printer_manager_log_loc =
      new LogicalLocality("printer_manager");

    // gateway connects to server
    gateway.addProcess(new SubscriberNodeCoordinator(server_loc,
        gateway_log_loc));
    gateway.waitToBeConnected();

    // printer manager connects to gateway
    printer_manager.addProcess(new SubscriberNodeCoordinator(gateway_loc,
        printer_manager_log_loc));
    printer_manager.waitToBeConnected();

    // printer connects to printer_manager
    printer.addProcess(new PrinterSubscriber(printer_manager_loc,
        printer_log_loc));
    printer.waitToBeConnected();

    // client connects to server
    printer_client.addProcess(new SubscriberNodeCoordinator(server_loc,
        new LogicalLocality("printer_client")));
    printer_client.waitToBeConnected();

    addProcessToNode(new PrinterServer(), printer);

    KlavaProcess printer_proc = new PrinterProcess();
    printer_proc.setNotifyOnNonExistentLocalities(true);
    addProcessToNode(printer_proc, printer_client);

    // printing on printer is OK
    printer_proc.join();

    // now printer disconnects
    printer.addProcess(new PrinterUnsubscriber(printer_manager_loc,
      printer_log_loc));

    KlavaProcess printer2_proc = new PrinterProcess("printer");
    printer2_proc.setKlavaProcName(printer2_proc.KlavaProcName() + "2");
    printer2_proc.setNotifyOnNonExistentLocalities(true);
    addProcessToNode(printer2_proc, printer_client);

    // printing on printer should now fail
    printer2_proc.join();
  }

  public static void main(String args[]) throws Exception {
    KlavaTestApplication app = new testPhysLocReach6
      ("testPhysLocReach6", (args.length > 1 ? args[1] : ""),
        args.length > 0);
    //app.start();
    app.waitForProcessesAndTerminate();
  }
}