package klavatest;

/**
 * test dispatching of messages in a hierarchical way
 *
 *                       server
 *                    ^           ^
 *          printer_manager      client_manager
 *        ^      ^       ^                    ^
 *    printer  printer2  printer3            client
 *
 * printer_client issues a tuple for printers and server has to
 * deliver it to printer, though printer is not directly connected to it.
 * One printer disconnects; then the printer_manager disconnects.
 * This tests the proxy removal for login/logout
 */

import Klava.*;

public class testPhysLocReachLogin3 extends KlavaTestApplication {
  public testPhysLocReachLogin3(String name, String suffix,
                                 boolean log_to_file)
                                 throws java.io.IOException,
                                 InterruptedException, KlavaException {
    super(name, suffix, log_to_file);

    NetNode server = new AcceptNetNode("server", 9999);
    server.setNoLogonServer(true); // this way it does not forward messages
    // to any other server

    NetNode printer_manager = new AcceptNetNode("printer_manager", 9998);

    NetNode client_manager = new AcceptNetNode("client_manager", 20000);
    NetNode printer_client = new NetNode("printer_client", 20001);

    NetNode printer = new NetNode("printer", 10001);
    NetNode printer2 = new NetNode("printer2", 10002);
    NetNode printer3 = new NetNode("printer3", 10003);

    PhysicalLocality server_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 9999));

    PhysicalLocality printer_manager_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 9998));

    PhysicalLocality printer_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 10001));
    PhysicalLocality printer2_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 10002));
    PhysicalLocality printer3_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 10003));

    PhysicalLocality client_manager_loc =
      new PhysicalLocality(NetUtils.createNodeAddress
        (NetUtils.getLocalIPAddress(), 20000));

    server.addToEnv("printer", printer_loc.toString());
    server.addToEnv("printer2", printer2_loc.toString());
    server.addToEnv("printer3", printer3_loc.toString());

    addNode(server);
    server.logRoutes("server");

    sortFiles("RoutingTable");

    server.start();
    printer_manager.start();
    printer.start();
    printer2.start();
    printer3.start();
    client_manager.start();
    printer_client.start();

    LogicalLocality printer_log_loc = new LogicalLocality("printer");
    LogicalLocality printer2_log_loc = new LogicalLocality("printer2");
    LogicalLocality printer3_log_loc = new LogicalLocality("printer3");

    printer.addProcess(new LoginNodeCoordinator(printer_manager_loc));
    printer.waitToBeConnected();

    printer2.addProcess(new LoginNodeCoordinator(printer_manager_loc));
    printer2.waitToBeConnected();

    printer3.addProcess(new LoginNodeCoordinator(printer_manager_loc));
    printer3.waitToBeConnected();

    LogicalLocality printer_manager_log_loc =
      new LogicalLocality("printer_manager");
    printer_manager.addProcess(new LoginNodeCoordinator(server_loc));
    printer_manager.waitToBeConnected();

    client_manager.addProcess(new LoginNodeCoordinator(server_loc));
    client_manager.waitToBeConnected();

    printer_client.addProcess(new LoginNodeCoordinator(client_manager_loc));
    printer_client.waitToBeConnected();

    addProcessToNode(new PrinterServer(), printer);
    KlavaProcess printer_proc = new PrinterProcess();
    printer_proc.setNotifyOnNonExistentLocalities(true);
    addProcessToNode(printer_proc, printer_client);

    // printing on printer is OK
    printer_proc.join();

    // now printer2 disconnects
    printer2.addProcess(new LogoutNodeCoordinator(printer_manager_loc));

    KlavaProcess printer2_proc = new PrinterProcess("printer2");
    printer2_proc.setKlavaProcName(printer2_proc.KlavaProcName() + "2");
    printer2_proc.setNotifyOnNonExistentLocalities(true);
    addProcessToNode(printer2_proc, printer_client);

    // printing on printer2 should fail
    printer2_proc.join();

    // now printer_manager disconnects
    LogoutNodeCoordinator logoutproc =
      new LogoutNodeCoordinator(server_loc);
    printer_manager.addProcess(logoutproc);
    logoutproc.join();

    KlavaProcess printer3_proc = new PrinterProcess("printer3");
    printer3_proc.setKlavaProcName(printer3_proc.KlavaProcName() + "3");
    printer3_proc.setNotifyOnNonExistentLocalities(true);
    addProcessToNode(printer3_proc, printer_client);

    // printing on printer3 should fail, since printer_manager disconnected
    printer3_proc.join();
  }

  public static void main(String args[]) throws Exception {
    KlavaTestApplication app = new testPhysLocReachLogin3
      ("testPhysLocReachLogin3", (args.length > 1 ? args[1] : ""),
        args.length > 0);
    //app.start();
    app.waitForProcessesAndTerminate();
  }
}