package org.mikado.imc.mobility;

/**
 * class providing method(s) for printing a message somewhere.  Implement a sort
 * of Decorator pattern.  It can be specialized in order to print to the
 * screen or to a file.
 *
 * @author $author$
 * @version $Revision: 1.2 $
 */
public class MessagePrinter {
    /** the name of this message printer */
    protected String name;

    /** the message printer that this one decorates */
    protected MessagePrinter decorated;

    /**
     * Creates a new MessagePrinter object.
     *
     * You can create a chain of message printers as in the following
     * example
     *
     * <pre>
     * MessagePrinter p =
     * new FileMessagePrinter("foo", new DefaultMessagePrinter(), "foo.out");
     * p.Print("hello world!");
     * </pre>
     *
     * this will print the message both into the file and to the screen
     *
     * @param name name for this message printer
     * @param decorated the message printer to decorate
     */
    public MessagePrinter(String name, MessagePrinter decorated) {
        this.name = name;
        this.decorated = decorated;
    }

    /**
     * print a message to the associated device and then delegate to
     * the decorate message printer.  The derived classes should always
     * call the super implementation so that the chain of decoration is
     * kept:
     *
     * <pre>
     * class MyMessagePrinter extends MessagePrinter {
     * ...
     *   public void Print(String s) {
     *      super.Print(s);
     *      // perform my cool printing
     *   }
     * }
     * </pre>
     *
     * @param s the string to print
     */
    public void Print(String s) {
        if (decorated != null) {
            decorated.Print(s);
        }
    }
}
